//------------------------------------------------------------------------------------------------+
// * @file  EnvironmentSimulator.cpp
// * @brief Simulate robot and people movement and provide connecting RTC sensor information
//------------------------------------------------------------------------------------------------+

#include "Simulator/Triangulation.h"
#include "EnvironmentSimulator.h"
#include "Simulator/Factory.h"


//------------------------------------------------------------------------------------------------+
// Module specification
//------------------------------------------------------------------------------------------------+
static const char* environmentsimulator_spec[] =
  {
    "implementation_id", "EnvironmentSimulator",
    "type_name",         "EnvironmentSimulator",
    "description",       "Simulate robot and people movement and provide connecting RTC sensor information",
    "version",           "1.0.0",
    "vendor",            "aisl.cs.tut.ac.jp",
    "category",          "Simulato",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.FieldDataFile", "Field.txt",
    "conf.default.robot_tread", "0.3325",
    "conf.default.robot_acceleration", "0.400",
    "conf.default.robot_deceleration", "0.400",
    "conf.default.robot_init_x", "3.5",
    "conf.default.robot_init_y", "1.0",
    "conf.default.robot_init_heading", "200.0",
    "conf.default.robot_move_error_x", "0.0",
    "conf.default.robot_move_error_y", "0.0",
    "conf.default.robot_move_error_heading", "0.0",
    "conf.default.walkingrate_mean", "1.33",
    "conf.default.walkingrate_std_deviation", "0.1",
    "conf.default.localmap_width", "200",
    "conf.default.localmap_height", "200",
    "conf.default.localmap_scale", "0.05",
    "conf.default.globalmap_scale", "0.1",
    "conf.default.show_map_scale", "0.05",
    "conf.default.following_target_id", "2",
    "conf.default.range_start_position", "-30.0",
    "conf.default.range_end_position", "210.0",
    "conf.default.range_max_distance", "30.0",
    "conf.default.range_far_code", "-1",
    "conf.default.range_data_num", "480",
    "conf.default.range_sensor_x", "0.0",
    "conf.default.range_sensor_y", "0.0",
    "conf.default.range_sensor_heading", "0.0",
    "conf.default.range_std_deviation", "0.0",
    "conf.default.range_person_radius", "0.2",
    "conf.default.display_range_data", "0",
    "conf.default.simulate_people", "1",
    "conf.default.simulate_skip", "1",
    // Widget
    "conf.__widget__.FieldDataFile", "text",
    "conf.__widget__.robot_tread", "text",
    "conf.__widget__.robot_acceleration", "text",
    "conf.__widget__.robot_deceleration", "text",
    "conf.__widget__.robot_init_x", "text",
    "conf.__widget__.robot_init_y", "text",
    "conf.__widget__.robot_init_heading", "text",
    "conf.__widget__.robot_move_error_x", "text",
    "conf.__widget__.robot_move_error_y", "text",
    "conf.__widget__.robot_move_error_heading", "text",
    "conf.__widget__.walkingrate_mean", "text",
    "conf.__widget__.walkingrate_std_deviation", "text",
    "conf.__widget__.localmap_width", "text",
    "conf.__widget__.localmap_height", "text",
    "conf.__widget__.localmap_scale", "text",
    "conf.__widget__.globalmap_scale", "text",
    "conf.__widget__.show_map_scale", "text",
    "conf.__widget__.following_target_id", "text",
    "conf.__widget__.range_start_position", "text",
    "conf.__widget__.range_end_position", "text",
    "conf.__widget__.range_max_distance", "text",
    "conf.__widget__.range_far_code", "text",
    "conf.__widget__.range_data_num", "text",
    "conf.__widget__.range_sensor_x", "text",
    "conf.__widget__.range_sensor_y", "text",
    "conf.__widget__.range_sensor_heading", "text",
    "conf.__widget__.range_std_deviation", "text",
    "conf.__widget__.range_person_radius", "text",
    "conf.__widget__.display_range_data", "text",
    "conf.__widget__.simulate_people", "text",
    "conf.__widget__.simulate_skip", "text",
    // Constraints
    ""
  };

//------------------------------------------------------------------------------------------------+
// constructor
//------------------------------------------------------------------------------------------------+
EnvironmentSimulator::EnvironmentSimulator(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_rangedataOut("rangedata", m_rangedata),
    m_robotControlInfoIn("robotControlInfo", m_robotControlInfo),
    m_robot_stateOut("robot_state", m_robot_state),
    m_robot_velocityOut("robot_velocity", m_robot_velocity),
    m_personsOut("persons", m_persons),
    m_peopledataOut("peopledata", m_peopledata),
    m_AbsoluteMapServicePort("AbsoluteMapService"),
    m_RelativeMapServicePort("RelativeMapService"),
    m_AreaInfoServicePort("AreaInfoService"),
	m_AbsoluteMapService( &globalmap, &cs_globalmap ),
	m_RelativeMapService( &localmap, &cs_localmap ),
    m_AreaInfoService( &subareas, &cs_subareas )
{

	// |C^ɈꉞNULLĂ
	sim       = 0;
	cTimer    = 0;
	window    = 0;

	Randomize();

	cerr << "------------------------------------------------------------\n"
		 << "  Environment Simulator Component (Ver. 3rd February 2012)\n"
		 << "------------------------------------------------------------\n\n";


	InitializeCriticalSection( &cs_localmap  );
	InitializeCriticalSection( &cs_globalmap );
	InitializeCriticalSection( &cs_subareas  );

	timeBeginPeriod( 1 );		// x1msɐݒ
}


//------------------------------------------------------------------------------------------------+
// destructor
//------------------------------------------------------------------------------------------------+
EnvironmentSimulator::~EnvironmentSimulator(){

	delete sim;
	delete cTimer;
	delete window;
}


//------------------------------------------------------------------------------------------------+
// onInitialize
//------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t EnvironmentSimulator::onInitialize(){

	// Set InPort buffers
	addInPort("robotControlInfo", m_robotControlInfoIn);

	// Set OutPort buffer
	addOutPort("rangedata", m_rangedataOut);
	addOutPort("robot_state", m_robot_stateOut);
	addOutPort("robot_velocity", m_robot_velocityOut);
	addOutPort("persons", m_personsOut);
	addOutPort("peopledata", m_peopledataOut);
  
	// Set service provider to Ports
	m_AbsoluteMapServicePort.registerProvider("AbsoluteMapService", "MRFC::AbsoluteMapService", m_AbsoluteMapService);
	m_RelativeMapServicePort.registerProvider("RelativeMapService", "MRFC::RelativeMapService", m_RelativeMapService);
	m_AreaInfoServicePort.registerProvider("AreaInfoService", "TUT::AreaInfoService", m_AreaInfoService);
	
	// Set CORBA Service Ports
	addPort(m_AbsoluteMapServicePort);
	addPort(m_RelativeMapServicePort);
	addPort(m_AreaInfoServicePort);

	// Bind variables and configuration variable
	bindParameter("FieldDataFile", m_FieldDataFile, "Field.txt");
	bindParameter("robot_tread", m_robot_tread, "0.3325");
	bindParameter("robot_acceleration", m_robot_acceleration, "0.400");
	bindParameter("robot_deceleration", m_robot_deceleration, "0.400");
	bindParameter("robot_init_x", m_robot_init_x, "3.5");
	bindParameter("robot_init_y", m_robot_init_y, "1.0");
	bindParameter("robot_init_heading", m_robot_init_heading, "200.0");
	bindParameter("robot_move_error_x", m_robot_move_error_x, "0.0");
	bindParameter("robot_move_error_y", m_robot_move_error_y, "0.0");
	bindParameter("robot_move_error_heading", m_robot_move_error_heading, "0.0");
	bindParameter("walkingrate_mean", m_walkingrate_mean, "1.33");
	bindParameter("walkingrate_std_deviation", m_walkingrate_std_deviation, "0.1");
	bindParameter("localmap_width", m_localmap_width, "200");
	bindParameter("localmap_height", m_localmap_height, "200");
	bindParameter("localmap_scale", m_localmap_scale, "0.05");
	bindParameter("globalmap_scale", m_globalmap_scale, "0.1");
	bindParameter("show_map_scale", m_show_map_scale, "0.05");
	bindParameter("following_target_id", m_following_target_id, "2");
	bindParameter("range_start_position", m_range_start_position, "-30.0");
	bindParameter("range_end_position", m_range_end_position, "210.0");
	bindParameter("range_max_distance", m_range_max_distance, "30.0");
	bindParameter("range_far_code", m_range_far_code, "-1");
	bindParameter("range_data_num", m_range_data_num, "480");
	bindParameter("range_sensor_x", m_range_sensor_x, "0.0");
	bindParameter("range_sensor_y", m_range_sensor_y, "0.0");
	bindParameter("range_sensor_heading", m_range_sensor_heading, "0.0");
	bindParameter("range_std_deviation", m_range_std_deviation, "0.0");
	bindParameter("range_person_radius", m_range_person_radius, "0.2");
	bindParameter("display_range_data", m_display_range_data, "0");
	bindParameter("simulate_people", m_simulate_people, "1");
	bindParameter("simulate_skip", m_simulate_skip, "1");

	return RTC::RTC_OK;
}


//------------------------------------------------------------------------------------------------+
// onActivated
//------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t EnvironmentSimulator::onActivated(RTC::UniqueId ec_id){

	cerr << "- Start Initialization -------------------------------------\n";

	step = 0;

	FieldFactory *fact = new FieldFactory( m_FieldDataFile.c_str() );
	field    = fact->field;
	actions  = fact->act_mng;

	subareas = SubAreas();

	if( field.EntranceNum() ){

		cerr << "\nStart to build spatial network.\n";
		Tessellator tessellator( field.SegmentsData() );
		subareas = tessellator.GenerateSubAreas();
		cerr << "\nNumber of generated Subareas = " << subareas.size() << endl;

		SubAreasManager area_mng( subareas );
		subareas = area_mng.ExtractInside( field.EntranceLine(0).Center() );
		cerr << "Number of optimised Subareas = " << subareas.size() << endl;
	}

	delete fact;
	fact = 0;

	robot = Robot( m_robot_init_x, m_robot_init_y, m_robot_init_heading );

	sensor = RangeSensor( m_range_start_position, m_range_end_position, m_range_max_distance, m_range_far_code, m_range_data_num, m_range_sensor_x, m_range_sensor_y, m_range_sensor_heading * deg_to_rad, m_range_std_deviation );

	// V~[^̏
	sim = new Simulator( &field, &actions, subareas, m_walkingrate_mean, m_walkingrate_std_deviation );
	sim->robots.push_back( robot );

	map = RoboMap( m_localmap_width, m_localmap_height, m_localmap_scale );
	MakeLocalMap();
	MakeGlobalMap();

	double scale = 1.0 / m_show_map_scale;
	window = new SimulationWindow( "Environment Simulator", (int)(field.FieldWidth()*scale)+1, (int)(field.FieldLength()*scale)+1, scale );


	odometry.SetParameter(m_robot_tread*1000., m_robot_acceleration*1000., m_robot_deceleration*1000., 0., 0., 0.);
	odometry.Reset( timeGetTime() );

	cerr << "\n- Finish Initialization ------------------------------------\n\n";

	// lo܂ŃXLbv
	if( m_simulate_skip && m_simulate_people && field.EntranceNum() ){

		while( sim->GetPeople().empty() || ( m_simulate_skip == 2 && 0 < m_following_target_id && sim->GetPeople().back().ID() != m_following_target_id )) sim->Simulation( SIMULATION_INTERVAL );
	}

	cTimer = new TIMER::TIMER_CLASS();		// ^C}[NXN
	cTimer->setTimer( T_TOTAL_TIME );		// vJn

	return RTC::RTC_OK;
}


//------------------------------------------------------------------------------------------------+
// onDeactivated
//------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t EnvironmentSimulator::onDeactivated(RTC::UniqueId ec_id){

	delete sim;
	sim = 0;

	// vI
	cTimer->stopTimer( T_TOTAL_TIME );

//	cout << "\n\nThe end of People Movement Simulation.\n\n"
	cout << "\n\nThe end of Simulation.\n\n"
		 << "Total Execution Time  : " << 0.001 * (double)cTimer->getValueAve(T_TOTAL_TIME)  << " [s]\n\n"
		 << "Required Time per cycle\n"
		 << "  Average : " <<  cTimer->getValueAve(T_ONE_CICLE) << " [ms/cycle]\n"
		 << "  Maximum : " <<  cTimer->getValueMax(T_ONE_CICLE) << " [ms]\n"
		 << "  Minimum : " <<  cTimer->getValueMin(T_ONE_CICLE) << " [ms]\n\n";

	delete cTimer;
	delete window;

	cTimer = 0;
	window = 0;

	return RTC::RTC_OK;
}


//------------------------------------------------------------------------------------------------+
// onExecute
//------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t EnvironmentSimulator::onExecute(RTC::UniqueId ec_id){

	cout << "\n[ STEP " << ++step << " ]\n";

	cTimer->setTimer( T_ONE_CICLE );

	// {bg̍XV
	UpdateRobotInfo();

	// lV~[^s
	if( m_simulate_people ) ExecuteSimulation();

	// ^CX^v̍XV
	time.sec  = timeGetTime() / 1000;
	time.nsec = ( timeGetTime() % 1000 ) * 1000000;

	// f[^̏o
	OutputRobotState();
	OutputRobotVelocity();
	OutputPersons();
	OutputPeopleData();
	OutputRangeData();

	// n}f[^̐
	MakeLocalMap();

	// \
	DrawWindow();

	cout << "Number of persons = " << sim->GetPeople().size() << endl;

	cTimer->stopTimer( T_ONE_CICLE );

	return RTC::RTC_OK;
}


//------------------------------------------------------------------------------------------------+
// V~[V̎s
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::ExecuteSimulation(){

	unsigned long now_time = timeGetTime();

	static unsigned long last_time = now_time - (unsigned long)( SIMULATION_INTERVAL * 1000 );

	// bȏ̃V~[V͂Ȃ
	if( 1000 < now_time - last_time ) last_time = now_time - 1000;

	double sim_interval = 0.001 * (double)( now_time - last_time );

	last_time = now_time;

	EnterCriticalSection( &cs_subareas );

	for( ; 0.0 < sim_interval; sim_interval -= SIMULATION_INTERVAL ){

		sim->Simulation(( sim_interval < SIMULATION_INTERVAL )? sim_interval: SIMULATION_INTERVAL );
	}

	LeaveCriticalSection( &cs_subareas );
}


//------------------------------------------------------------------------------------------------+
// {bg̃Ihgo
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::OutputRobotState(){

	m_robot_state.data.position.x = robot.x;	// odometry.GetPositionX()/1000.;
	m_robot_state.data.position.y = robot.y;	// odometry.GetPositionY()/1000.;
	m_robot_state.data.heading = robot.theta;	// odometry.GetTheta();

	m_robot_stateOut.write( m_robot_state );
}


//------------------------------------------------------------------------------------------------+
// {bg̑xo
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::OutputRobotVelocity(){

	m_robot_velocity.data.vx = odometry.GetVelocityV()/1000.;
	m_robot_velocity.data.vy = 0.0;
	m_robot_velocity.data.va = odometry.GetVelocityW();

	m_robot_velocityOut.write( m_robot_velocity );
}


//------------------------------------------------------------------------------------------------+
// iWnjlʒuf[^̏o
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::OutputPersons(){

	m_persons.tm = time;

	m_persons.data.length( sim->GetPeople().size() );

	for( int i=0; i< (int)sim->GetPeople().size(); i++ ){

		m_persons.data[i].x  = sim->GetPeople()[i].p.x;
		m_persons.data[i].y  = sim->GetPeople()[i].p.y;
	}

	m_personsOut.write( m_persons );
}


//------------------------------------------------------------------------------------------------+
// iǏWnjlf[^̏o
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::OutputPeopleData(){

	m_peopledata.tm = time;

	m_peopledata.data.person.length( sim->GetPeople().size() );
	m_peopledata.data.id = -1;

	for( int i=0; i< (int)sim->GetPeople().size(); i++ ){

		if( sim->GetPeople()[i].ID() == m_following_target_id ) m_peopledata.data.id = i;

		// l̃xNgǏWnɕϊ
		Segment person = map.Convert( sim->GetPeople()[i].p, &robot );

		m_peopledata.data.person[i].position.x  = person.x;
		m_peopledata.data.person[i].position.y  = person.y;
		m_peopledata.data.person[i].position.z  = 0.0;
		m_peopledata.data.person[i].velocity.vx = person.dx;
		m_peopledata.data.person[i].velocity.vy = person.dy;
		m_peopledata.data.person[i].velocity.va = 0.0;
	}

	m_peopledataOut.write( m_peopledata );
}


//------------------------------------------------------------------------------------------------+
// f[^̎擾Əo
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::OutputRangeData(){

	sensor.Sensing( &robot, &field, &(sim->GetPeople()), m_range_person_radius );

	m_rangedata.data.dataGroupingNumber = 1;
	m_rangedata.data.dataInterval       = sensor.DataInterval();
	m_rangedata.data.startPosition      = sensor.start_position;
	m_rangedata.data.endPosition        = sensor.end_position;
	m_rangedata.data.sensorState        = "NORMAL";

	m_rangedata.tm = time;

	m_rangedata.data.distance.length( sensor.data_num );

	cerr << endl;

	for( int i=0; i<sensor.data_num; i++ ) m_rangedata.data.distance[i] = sensor.range_data[i];

	cerr << endl;

	m_rangedataOut.write();
}


//------------------------------------------------------------------------------------------------+
// n}̐
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::MakeGlobalMap(){

	EnterCriticalSection( &cs_globalmap );

	globalmap = TimedAbsoluteOGMapData();

	globalmap.mapconfig.width  = (long)( field.FieldWidth()  / m_globalmap_scale ) +1;
	globalmap.mapconfig.height = (long)( field.FieldLength() / m_globalmap_scale ) +1;
	globalmap.mapconfig.xScale = m_globalmap_scale;
	globalmap.mapconfig.yScale = m_globalmap_scale;

	globalmap.mapconfig.origin.heading    = 0.0;
	globalmap.mapconfig.origin.position.x = 0.0;
	globalmap.mapconfig.origin.position.y = 0.0;

	globalmap.tm = time;

	RoboMap g_robomap( globalmap.mapconfig.width, globalmap.mapconfig.height, m_globalmap_scale );

	Robot dummy_robo = Robot( field.FieldWidth()/2.0, field.FieldLength()/2.0, 0.0 );

	g_robomap.MakeMap( &field, &dummy_robo );

	globalmap.cells.length( globalmap.mapconfig.width * globalmap.mapconfig.height );

	int index = 0;

	for( int y=0; y<globalmap.mapconfig.height; y++ ) for( int x=0; x<globalmap.mapconfig.width; x++ ){

		globalmap.cells[index] = (g_robomap.map[y][x])? 100: 0;
		index++;
	}

	LeaveCriticalSection( &cs_globalmap );
}


//------------------------------------------------------------------------------------------------+
// Ǐn}̐
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::MakeLocalMap(){

	EnterCriticalSection( &cs_localmap );

	localmap = TimedRelativeOGMapData();

	localmap.mapconfig.width  = m_localmap_width;
	localmap.mapconfig.height = m_localmap_height;
	localmap.mapconfig.xScale = m_localmap_scale;
	localmap.mapconfig.yScale = m_localmap_scale;

	localmap.mapconfig.origin.heading    = 0.0;
	localmap.mapconfig.origin.position.x = -m_localmap_scale * (double)m_localmap_width  / 2.0;
	localmap.mapconfig.origin.position.y = -m_localmap_scale * (double)m_localmap_height / 2.0;

	localmap.tm = time;

	localmap.pose.position.x = robot.x;
	localmap.pose.position.y = robot.y;
	localmap.pose.heading    = robot.theta;

	localmap.cells.length( m_localmap_width * m_localmap_height );

	int index = 0;

	for( int y=0; y<m_localmap_height; y++ ) for( int x=0; x<m_localmap_width; x++ ){

		localmap.cells[index] = (map.map[y][x])? 100: 0;
		index++;
	}

	LeaveCriticalSection( &cs_localmap );
}


//------------------------------------------------------------------------------------------------+
// {bg̍XV
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::UpdateRobotInfo(){

	if( m_robotControlInfoIn.isNew() ){

		m_robotControlInfoIn.read();
		odometry.ControlReceive(m_robotControlInfo.data.vx*1000., m_robotControlInfo.data.va, timeGetTime());
	}
	else odometry.Update( timeGetTime() );		// Ihg̉Zs(䖽߂MĂȂꍇ)

	double robot_x, robot_y, robot_theta;

	double cos_value = cos( m_robot_init_heading * deg_to_rad );
	double sin_value = sin( m_robot_init_heading * deg_to_rad );

	robot_x = odometry.GetPositionX() * cos_value - odometry.GetPositionY() * sin_value;
	robot_y = odometry.GetPositionX() * sin_value + odometry.GetPositionY() * cos_value;

	robot_x /= 1000.0;
	robot_y /= 1000.0;

	robot_x += m_robot_init_x;	// 摜E 
	robot_y += m_robot_init_y;	// 摜 
	robot_theta = odometry.GetTheta()+m_robot_init_heading * deg_to_rad;	// E]{,摜E0 

	// 덷
	double dx = robot_x - robot.x;
	double dy = robot_y - robot.y;
	double dist = sqrt( dx * dx + dy * dy );

	double error_x = dist * GausRand( 0.0, m_robot_move_error_x ).Rand();
	double error_y = dist * GausRand( 0.0, m_robot_move_error_y ).Rand();

	robot_x     += error_x * cos_value - error_y * sin_value;
	robot_y     += error_x * sin_value + error_y * cos_value;
	robot_theta += dist * GausRand( 0.0, m_robot_move_error_heading * deg_to_rad ).Rand();

	sim->robots[0] = robot = Robot( robot_x, robot_y, robot_theta );

	map.MakeMap( &field, &robot );		// n}̍XV
}


//------------------------------------------------------------------------------------------------+
// EBhE̕`
//------------------------------------------------------------------------------------------------+
void EnvironmentSimulator::DrawWindow(){

	window->Reset();

	if( m_display_range_data ){
		for( int i=0; i<sensor.data_num; i++ ) window->DrowLine( sensor.rays[i], 0, 0, 255 );	// [U
	}

	for( int i=0; i<field.WallNum(); i++ ) window->DrowWall( field.Wall(i) );							// ǂ\

	for( int i=0; i<(int)sim->robots.size(); i++ ){

		Line robo = Line( sim->robots[i].x, sim->robots[i].y, cos( sim->robots[i].theta ), sin( sim->robots[i].theta ));
		window->DrowPerson( robo, 255, 127 );		// {bg\
	}

	for( int i=0; i<(int)sim->GetPeople().size(); i++ ){

		if( sim->GetPeople()[i].ID() == m_following_target_id )  window->DrowPerson( (Line)sim->GetPeople()[i].p, 255, 0 );	// l\
		else window->DrowPerson( (Line)sim->GetPeople()[i].p );	// l\
	}

	window->Show();
}



//------------------------------------------------------------------------------------------------+
extern "C"
{
 
  void EnvironmentSimulatorInit(RTC::Manager* manager)
  {
    coil::Properties profile(environmentsimulator_spec);
    manager->registerFactory(profile,
                             RTC::Create<EnvironmentSimulator>,
                             RTC::Delete<EnvironmentSimulator>);
  }
};
